import timing
import compiler
import Numeric
import numarray
import numarray._numarray as _numarray
import numarray.memory as memory
import numarray._ufunc as _ufunc
import sys
import string
import re
import profile
import hotshot
import hotshot.stats

parameters = """
i=0 a=arange(10**i) j=int(3) b=reshape(arange(j*j),(j,j))           c=a.copy() d=a.copy() e=transpose(b) f=transpose(b) af=a.astype('d') cf=a.astype('d') z=array(0)
i=1 a=arange(10**i) j=int(3) b=reshape(arange(j*j),(j,j))           c=a.copy() d=a.copy()
i=2 a=arange(10**i) j=int(4) b=reshape(arange(j*j),(j,j))           c=a.copy() d=a.copy()
i=3 a=arange(10**i) j=int(10.0**1.5) b=reshape(arange(j*j),(j,j))   c=a.copy() d=a.copy()
i=4 a=arange(10**i) j=int(10.0**2.0) b=reshape(arange(j*j),(j,j))   c=a.copy() d=a.copy()
i=5 a=arange(10**i) j=int(10.0**2.5) b=reshape(arange(j*j),(j,j))   c=a.copy() d=a.copy()
i=6 a=arange(10**i) j=int(10.0**3) b=reshape(arange(j*j),(j,j))   c=a.copy() d=a.copy() e=transpose(b) f=transpose(b) af=a.astype('d') cf=a.astype('d') z=array(0)
"""

benchmarks = """
NumArray(buffer=None,shape=(10**1,),type=Int32)
NewArray(buffer=None,shape=(10**1,),type=Int32)
a.tolist()
a.tostring()
array(a)
setitem(a)
getitem(a)
add(a,c)          Int32,Int32        fast case implicit output
add(a,1)          Int32,scalar       fast case implicit output
add(a,1.0)        Int32,scalar       slow case implicit output, type converted
add(a,c,d)        Int32,Int32,Int32  fast case explicit output
add(e,f)          Int32,Int32        slow case discontiguous implicit output
add(e,f,b)        Int32,Int32,Int32  slow case discontiguous explicit output
add(a,af)         Int32,Float64      slow case type converted
add(b,b)          Int32,Int32        two dimensional
add.reduce(b)     Int32              fast case implicit output swapaxes
add.reduce(b,-1)  Int32              fast case implicit output 
add.reduce(a,-1,z) Int32             fast case explicit output
add.reduce(e) Int32                  slow case implicit output
add.reduce(e,-1,e) Int32, Int32      slow case explicit output w/conversion
add.accumulate(a) Int32              fast case implicit output swapaxes
add.accumulate(a,-1) Int32           fast case implicit output 
add.accumulate(a,-1,a) Int32         fast case explicit output
add.accumulate(e) Int32              slow case implicit output
add.accumulate(e,-1,e) Int32, Int32  slow case explicit output w/conversion
add.areduce(b)     Int32             fast case implicit output swapaxes
add.areduce(b,-1)  Int32             fast case implicit output 
add.areduce(a,-1,z) Int32            fast case explicit output
add.areduce(e) Int32                 slow case implicit output
add.areduce(e,-1,e) Int32, Int32     slow case explicit output w/conversion
cos(af)           Float64            fast case
cos(1.0)          scalar             fast case
cos(1)            scalar             slow case type converted
cos(e)            Int32              slow case discontiguous, type converted
cos(af,cf)                           fast case explicit output
take(a,c)
put(a,c,d)
nonzero(a%2)
argsort(a)
argsort(b)
sort(a)
sort(b)
a[0]
a[::2]
a[::2] = c[::2]
matrixmultiply(b,b)
searchsorted(a,a)
concatenate((a,)*2)
concatenate((a,)*5)
resize(a,(10**i/2,))
resize(a,(10**i*2,))
compress(a%2,a)
transpose(b)
swapaxes(b,0,-1)
diagonal(b)
trace(b)
where(a%2,a,a)
indices((j,))
indices((j,j))
putmask(a,c%2,d)
clock()
dummy(1,2,1,((1,1),(2,2),(3,3)))
new_memory(10**6)
new_memory(1)
"""

def setitem(a):
    for i in xrange(len(a)):
        a[i] = 0

def getitem(a):
    for i in xrange(len(a)):
        j = a[i]

def dummy(*args):
    pass

MIN_TIME_RESOLUTION = 0.100

def otime(gs, ls, N):
    timing.start()
    for i in range(N):
        pass
    timing.finish()
    return timing.micro()
    
def time(bmark, gs, ls, N=1, force_exit=0):
    code = compiler.compile(bmark,"benchmarks","eval")
    try:
        timing.start()
        for i in range(N):
            eval(code, ls, gs)
        timing.finish()
        t = timing.micro()/10.0**6
        if t > MIN_TIME_RESOLUTION or force_exit:
            return t/N  # (t - otime(gs, ls, N))/N
        else:
            return time(bmark, gs, ls, N*2)
    except:
        return None

def run(num, bmark, p):
    locals = {}
    locals["setitem"] = setitem
    locals["getitem"] = getitem
    locals["dummy"] = dummy
    locals["new_memory"] = memory.new_memory
    try:
        locals["digest"] = _numarray.digest
    except:
        pass 
    for i in string.split(p," "):
        exec i in num, locals
    return time(bmark, locals, num)

def pp(results):
    s = "("
    for i in range(len(results)):
        if results[i] != None:
            s += ("%4.2f, " % results[i])
        else:
            s += "None, "
    return s[:-2] + ")"

def case(line):
    results = []
    params = parameters.split("\n")
    bmark = line.split(" ")[0]
    for i in range(len(params)):
        p = params[i]
        if len(p):
            m0 = run(Numeric.__dict__, bmark, p)
            n0 = run(numarray.__dict__, bmark, p)
            if m0 and n0:
                print "Numeric:", m0, "numarray:", n0
                results.append(m0/n0)
            else:
                results.append(None)
    print line, ":", pp(results)

"""
vector scalar Float32/Float32 add
add(a1,1.0)
   Overhead in us [ratio]    per element time (ns) [ratio]
        200       [22.1]          80               [0.5]

"""

def report(line):
    params = parameters.split("\n")
    p0 = params[1] # 10**0 element case
    p6 = params[7] # 10**6 element case

    bmark = line.split(" ")[0]
    
    m0 = run(Numeric.__dict__, bmark, p0)
    n0 = run(numarray.__dict__, bmark, p0)
    m6 = run(Numeric.__dict__, bmark, p6)
    n6 = run(numarray.__dict__, bmark, p6)

    if m0 and m0:
        ne = ((n6-n0)/10**6*10**9)
        me = ((m6-m0)/10**6*10**9)
        print line
        print "%20s%8s%20s%8s" % ("overhead (usec)","[ratio]","per element (nsec)","[ratio]")
        try:
            r1 = n0/m0
        except:
            r1 = 0
        try:
            r2 = ne/me
        except:
            r2 = 0
        print "%20d%8.1f%20d%8.1f" % (n0*10**6, r1, ne, r2)
    elif n0:
        ne = ((n6-n0)/10**6*10**9)
        print bmark
        print "%20s%8s%20s%8s" % ("overhead (usec)","[ratio]","per element (nsec)","[ratio]")
        print "%20d%8s%20d%8s" % (n0*10**6, None, ne, None)


def marks(pattern):
    if pattern != None:
        l = []
        lines = string.split(benchmarks,"\n")
        for i in range(len(lines)):
            if re.match(pattern, lines[i]):
                l.append(lines[i])
    else:
        l = benchmarks.split("\n")
    return l

def benchmark(pattern=None, f=report):
    for m in marks(pattern):
        if m:
            f(m)

def prof_do(mark, m, verbose=1):
    pr = hotshot.Profile("prof.out", 1, 1)
    # pr.calibrate(100, verbose=1)
    pr.run(mark)
    if verbose:
        print "profile for: ", m
    pr.close()
    pr = hotshot.stats.load("prof.out")
    pr.strip_dirs().sort_stats("time")
    pr.print_stats()

def prof(m, verbose=1):
    params = parameters.split("\n")
    globals().update(numarray.__dict__)
    for p in params[1].split():
        exec p in globals()
    line = m.split(" ")
    mark, comment = line[0], " ".join(line[1:])
    prof_do(mark, m, verbose)

def profall(pattern=None):
    for m in marks(pattern):
        if m:
            print "-"*70
            prof(m)
            prof(m + " (cached)")

def joint(pattern=None):
    for m in marks(pattern):
        if m:
            print "-"*70
            report(m)
            prof(m + "(cached)", verbose=0)

def exercise(m, N=100000):
    params = parameters.split("\n")
    globals().update(numarray.__dict__)
    for p in params[1].split():
        exec p in globals()
    line = m.split(" ")
    mark, comment = line[0], " ".join(line[1:])
    code = compiler.compile(mark,"exercise","eval")
    print '='*70
    print m
    print '='*70
    for i in xrange(N):
        eval(code, globals())

def gprof(pattern=None):
    for m in marks(pattern):
        if m:
            exercise(m)

if __name__ == "__main__":
    if len(sys.argv) == 1:
        benchmark()
    else:
        if sys.argv[1] == "prof":
            for p in sys.argv[1:]:
                profall(p)
        elif sys.argv[1] == "joint":
            for p in sys.argv[1:]:
                joint(p)
        elif sys.argv[1] == "gprof":
            gprof(sys.argv[2])
        elif "times" == sys.argv[1]:
            for b in sys.argv[1:]:
                benchmark(b, f=case)
        else:
            for b in sys.argv[1:]:
                benchmark(b, f=report)
