# numarray convolve-example tests
#
# Do not run this file directly, it is used from convtest.py

import numarray as num
import sys
import unittest

array = num.array
Float64 = num.Float64
Int32 = num.Int32


# some helper functions used in the tests
def dotest(f, k, x, z=None):
    kref = sys.getrefcount(k)
    xref = sys.getrefcount(x)
    zref = sys.getrefcount(z)
    if z is None:
        c = f(k, x)
    else:
        c = f(k, x, z)
    assert(sys.getrefcount(k) == kref)
    assert(sys.getrefcount(x) == xref)
    if z is None:
        assert(sys.getrefcount(c) == 2)
    else:
        assert(sys.getrefcount(z) == zref)
    if c is not None:
        return c
    else:
        return z

def conv1(k, x, z=None):
    return dotest(convolve.Convolve1d, k, x, z)
    
def conv2(k, x, z=None):
    return dotest(convolve.Convolve2d, k, x, z)

def test():
    import doctest, convtest
    return doctest.testmod(convtest)



def logical_reduce(a):
    return num.logical_and.reduce(num.ravel(a))

# actual test-class

class convolve_test(unittest.TestCase):

    def _FailIf(self, x):
        return self.failIf( logical_reduce(x) )

    def test_001(self):
        """Unconverted data, unconverted output."""
        self._FailIf(
            conv1(num.array([1.,2.,1.]), num.arange(10.),
                  num.zeros((10,), type=Float64)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))

    def test_002(self):
        """Converted data, unconverted output."""
        self._FailIf(
            conv1(num.array([1.,2.,1.]), num.arange(10),
                  num.zeros((10,), type=Float64)).__ne__(
            array([0, 4, 8, 12, 16, 20, 24, 28, 32, 9])))

    def test_003(self):
        """Unconverted data, converted output."""
        self._FailIf(
            conv1(num.array([1.,2.,1.]), num.arange(10.),
                  num.zeros((10,), type=Int32)).__ne__(
            array([0, 4, 8, 12, 16, 20, 24, 28, 32, 9])))

    def test_004(self):
        """Converted data, converted output."""
        self._FailIf(
            conv1(num.array([1.,2.,1.]), num.arange(10.),
                  num.zeros((10,), type=Int32)).__ne__(
            array([0, 4, 8, 12, 16, 20, 24, 28, 32, 9])))

    def test_005(self):
        """Converted data, no output."""
        self._FailIf(conv1(num.array([1.,2.,1.]), num.arange(10.)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))

    def test_006(self):
        """Unonverted data, no output."""
        self._FailIf(
            conv1(num.array([1.,2.,1.]), num.arange(10)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))

    def test_007(self):
        """List data, no output."""
        self._FailIf(
            conv1([1.,2.,1.], num.arange(10)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))

    def test_008(self):
        """Byteswapped data, no output."""
        a = num.array([1.,2.,1.])
        a.byteswap()
        self._FailIf(
            conv1(a, range(10)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))
        
    def test_008(self):
        """Misaligned data, no output."""
        a = num.array([1.,2.,1.])
        a._aligned = 1
        self._FailIf(conv1(a, range(10)).__ne__(
            array([0., 4., 8., 12., 16., 20., 24., 28., 32., 9.])))
        
    def test_101(self):
        """2d convolve 10x10 by 5,5."""
        x = num.zeros((10,10), num.Float64) + 1
        k = num.zeros((5,5), num.Float64) + 0.25
        k[1:4,1:4] = 0.5
        k[2,2] = 1
        z = num.zeros((10,10), num.Float64)
        self._FailIf(conv2(k, x, z).__ne__(
            array([[ 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                   [ 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 9., 9., 9., 9., 9., 9., 1., 1.],
                   [ 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                   [ 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.]])))
                    


class NoExitTestProgram(unittest.TestProgram):
    """A command-line program that runs a set of tests; this is primarily
       for making test modules conveniently executable.

       Does not exit explicitely!
    """
    def runTests(self):
        if self.testRunner is None:
            self.testRunner = TextTestRunner(verbosity=self.verbosity)
        result = self.testRunner.run(self.test)
        return(not result.wasSuccessful())

from unittest import *
NoExitTestProgram()
